//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "mpu680x0.h"

// 現在の FC での通常アクセス (MPU 内部から使用する)
// ミスアライン可。
// バスエラーが起きると例外 M68K::EXCEP_BUSERR をスローする。
//	uint32 fetch_{16,32}()
//	uint32 read_{8,16,32}(laddr)
//	void   write_{8,16,32}(laddr, data)

// reg.pc からワードをフェッチし reg.pc を進める。
uint32
MPU680x0Device::fetch_16()
{
	uint32 data;

	if (__predict_true(icache_enable)) {
		data = fetch_16_cache(reg.pc);
	} else {
		data = fetch_16_bus(reg.pc);
	}
	reg.pc += 2;

	return data;
}

// reg.pc からロングワード(ミスアライン可)をフェッチし reg.pc を進める。
uint32
MPU680x0Device::fetch_32()
{
	uint32 data;

	if (__predict_true(icache_enable)) {
		data = fetch_32_cache(reg.pc);
	} else {
		if ((reg.pc & 2) == 0) {
			data = fetch_32_bus(reg.pc);
		} else {
			data  = fetch_16_bus(reg.pc) << 16;
			data |= fetch_16_bus(reg.pc + 2);
		}
	}
	reg.pc += 4;

	return data;
}

uint32
MPU680x0Device::read_8(uint32 addr)
{
	busaddr laddr(addr, fc_data | busaddr::R);
	return read_8_fc(laddr);
}

uint32
MPU680x0Device::read_16(uint32 addr)
{
	busaddr laddr(addr, fc_data | busaddr::R);
	return read_16_fc(laddr);
}

uint32
MPU680x0Device::read_32(uint32 addr)
{
	busaddr laddr(addr, fc_data | busaddr::R);
	return read_32_fc(laddr);
}

void
MPU680x0Device::write_8(uint32 addr, uint32 data)
{
	busaddr laddr(addr, fc_data | busaddr::W);
	write_8_fc(laddr, data);
}

void
MPU680x0Device::write_16(uint32 addr, uint32 data)
{
	busaddr laddr(addr, fc_data | busaddr::W);
	write_16_fc(laddr, data);
}

void
MPU680x0Device::write_32(uint32 addr, uint32 data)
{
	busaddr laddr(addr, fc_data | busaddr::W);
	write_32_fc(laddr, data);
}

// FC 指定アクセス (MPU コアから使用する)
// ミスアライン可。
// バスエラーが起きると例外 M68K::EXCEP_BUSERR をスローする。
//	uint32 read_{8,16,32}_fc(busaddr)
//	void   write_{8,16,32}_fc(busaddr, data)
// (fetch は常に現在の FC でアクセスするので FC 指定版は不要)

uint32
MPU680x0Device::read_8_fc(busaddr addr)
{
	uint32 data;

	if (1) {
		data = read_8_bus(addr);
	}

	return data;
}

uint32
MPU680x0Device::read_16_fc(busaddr addr)
{
	uint32 data;

	if (1) {
		// バスアクセスは要整列。
		if (__predict_true((addr & 1) == 0)) {
			data = read_16_bus(addr);
		} else {
			data  = read_8_bus(addr) << 8;
			addr += 1;
			data |= read_8_bus(addr);
		}
	}

	return data;
}

uint32
MPU680x0Device::read_32_fc(busaddr addr)
{
	uint32 data;

	if (1) {
		// バスアクセスは要整列。
		if (__predict_true((addr & 3) == 0)) {
			data = read_32_bus(addr);
		} else if (__predict_true((addr & 1) == 0)) {
			data  = read_16_bus(addr) << 16;
			addr += 2;
			data |= read_16_bus(addr);
		} else {
			data  = read_8_bus(addr) << 24;
			addr += 1;
			data |= read_16_bus(addr) << 8;
			addr += 2;
			data |= read_8_bus(addr);
		}
	}

	return data;
}

void
MPU680x0Device::write_8_fc(busaddr addr, uint32 data)
{
	if (1) {
		write_8_bus(addr, data);
	}
}

void
MPU680x0Device::write_16_fc(busaddr addr, uint32 data)
{
	if (1) {
		// バスアクセスは要整列。
		if (__predict_true((addr & 1) == 0)) {
			write_16_bus(addr, data);
		} else {
			write_8_bus(addr, data >> 8);
			addr += 1;
			write_8_bus(addr, data & 0xff);
		}
	}
}

void
MPU680x0Device::write_32_fc(busaddr addr, uint32 data)
{
	if (1) {
		// バスアクセスは要整列。
		if (__predict_true((addr & 3) == 0)) {
			write_32_bus(addr, data);
		} else if (__predict_true((addr & 1) == 0)) {
			write_16_bus(addr, data >> 16);
			addr += 2;
			write_16_bus(addr, data & 0xffff);
		} else {
			write_8_bus(addr, data >> 24);
			addr += 1;
			write_16_bus(addr, (data >> 8) & 0xffff);
			addr += 2;
			write_8_bus(addr, data & 0xff);
		}
	}
}

// 外部バスアクセス (MMU を含む)
// ミスアライン不可。
// バスエラーが起きると例外 M68K::EXCEP_BUSERR をスローする。
//	uint32 fetch_{16,32}_bus()
//	uint32 read_{8,16,32}_bus(busaddr)
//	void   write_{8,16,32}_bus(busaddr, data)

// goto を含むことに注意。gcc 拡張式(?) に注意。
#define READ(size)	({	\
	bus.paddr = bus.laddr;	\
	bus.ci = false;	\
	if (__predict_false(TranslateRead() == false))	\
		goto buserr;	\
	busdata bd_ = mainbus->__CONCAT(Read,size)(bus.paddr);	\
	buswait += bd_.GetWait();	\
	if (bd_.IsBusErr()) \
		goto buserr;	\
	bd_;	/* これが戻り値 */	\
})
#define WRITE(size, data_)	({	\
	bus.paddr = bus.laddr;	\
	bus.ci = false;	\
	if (__predict_false(TranslateWrite() == false))	\
		goto buserr;	\
	busdata bd_ = mainbus->__CONCAT(Write,size)(bus.paddr, data_); \
	buswait += bd_.GetWait();	\
	if (bd_.IsBusErr()) \
		goto buserr;	\
	bd_;	/* これが戻り値 */	\
})

uint32
MPU680x0Device::fetch_16_bus(uint32 addr)
{
	busdata bd;

	bus.laddr.Set(addr, fc_inst);
	bd = READ(16);
	return bd.Data();

 buserr:
	bus.size = M68K::SSW_SIZE_WORD;
	throw M68K::EXCEP_BUSERR;
}

uint32
MPU680x0Device::fetch_32_bus(uint32 addr)
{
	busdata bd;

	bus.laddr.Set(addr, fc_inst);
	bd = READ(32);
	return bd.Data();

 buserr:
	bus.size = M68K::SSW_SIZE_LONG;
	throw M68K::EXCEP_BUSERR;
}

uint32
MPU680x0Device::read_8_bus(busaddr addr)
{
	busdata data;

	bus.laddr = addr;
	data = READ(8);
	putlog(2, "read_8   %08x -> %02x", bus.laddr.Addr(), data.Data());
	return data.Data();

 buserr:
	bus.size = M68K::SSW_SIZE_BYTE;
	throw M68K::EXCEP_BUSERR;
}

uint32
MPU680x0Device::read_16_bus(busaddr addr)
{
	busdata data;

	bus.laddr = addr;
	data = READ(16);
	putlog(2, "read_16  %08x -> %04x", bus.laddr.Addr(), data.Data());
	return data.Data();

 buserr:
	bus.size = M68K::SSW_SIZE_WORD;
	throw M68K::EXCEP_BUSERR;
}

uint32
MPU680x0Device::read_32_bus(busaddr addr)
{
	busdata data;

	bus.laddr = addr;
	data = READ(32);
	putlog(2, "read_32  %08x -> %08x", bus.laddr.Addr(), data.Data());
	return data.Data();

 buserr:
	bus.size = M68K::SSW_SIZE_LONG;
	throw M68K::EXCEP_BUSERR;
}

void
MPU680x0Device::write_8_bus(busaddr addr, uint32 data)
{
	busdata rd;

	bus.laddr = addr;
	rd = WRITE(8, data);
	putlog(2, "write_8  %08x <- %02x", bus.laddr.Addr(), data);
	return;

 buserr:
	bus.size = M68K::SSW_SIZE_BYTE;
	throw M68K::EXCEP_BUSERR;
}

void
MPU680x0Device::write_16_bus(busaddr addr, uint32 data)
{
	busdata rd;

	bus.laddr = addr;
	rd = WRITE(16, data);
	putlog(2, "write_16 %08x <- %04x", bus.laddr.Addr(), data);
	return;

 buserr:
	bus.size = M68K::SSW_SIZE_WORD;
	throw M68K::EXCEP_BUSERR;
}

void
MPU680x0Device::write_32_bus(busaddr addr, uint32 data)
{
	busdata rd;

	bus.laddr = addr;
	rd = WRITE(32, data);
	putlog(2, "write_32 %08x <- %08x", bus.laddr.Addr(), data);
	return;

 buserr:
	bus.size = M68K::SSW_SIZE_LONG;
	throw M68K::EXCEP_BUSERR;
}

// バースト転送。
// addr の下位4ビットは無視する。
// addr がバースト転送可能なら dst に 16 バイト転送して true を返す。
// addr がバースト転送不可なら dst は触らず false を返す。
// VM (バス) がバースト転送対応している場合のみ呼ぶこと。
bool
MPU680x0Device::fetch_burst_bus(uint32 addr, uint32 *dst)
{
	bus.laddr.Set(addr & 0xfffffff0, fc_inst);
	bus.paddr = bus.laddr;
	bus.ci = false;
	if (__predict_false(TranslateRead() == false)) {
		bus.size = M68K::SSW_SIZE_LONG;
		throw M68K::EXCEP_BUSERR;
	}
	if (__predict_false(bus.ci)) {
		return false;
	}
	busdata bd = mainbus->ReadBurst16(bus.paddr, dst);
	buswait += bd.GetWait();

	// BusErr ビットをバースト転送不可フラグとして使っている。
	return !bd.IsBusErr();
}
