//
// nono
// Copyright (C) 2023 nono project
// Licensed under nono-license.txt
//

//
// Nereid バンクメモリ
//

// バンクメモリはロングワード単位でホストバイトオーダ配置なので、
// バイトアクセス、ワードアクセス時は HLB(), HLW() マクロを使用のこと。

#include "bankram.h"
#include "config.h"
#include "mainbus.h"

// コンストラクタ
BankRAMDevice::BankRAMDevice(int objid_, int ram_size_MB)
	: inherited(objid_)
{
	// ram_size_MB は正数のみ。負数は親側で処理する。
	assert(ram_size_MB >= 0);

	ram_size = ram_size_MB * 1024 * 1024;
	// 初期化は Init() で行う
}

// デストラクタ
BankRAMDevice::~BankRAMDevice()
{
}

// 初期化
bool
BankRAMDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	ram.reset(new uint8[ram_size]);

	return true;
}

// 電源オン/リセット
void
BankRAMDevice::ResetHard(bool poweron)
{
	if (poweron) {
		// ノイズで埋めたいがとりあえず。
		memset(&ram[0], 0, ram_size);
	}

	// リセットでクリアされる。
	SetPage(0);

	std::fill(accstat_read.begin(), accstat_read.end(), 0);
	std::fill(accstat_write.begin(), accstat_write.end(), 0);
}

inline uint32
BankRAMDevice::Decoder(uint32 addr) const
{
	return pageaddr + (addr & 0xffff);
}

busdata
BankRAMDevice::Read8(uint32 addr)
{
	uint32 offset = Decoder(addr);
	accstat_read[offset >> 20] = MainbusDevice::ACC_READ;
	uint32 data = ram[HLB(offset)];
	putlog(4, "$%08x.B -> $%02x", addr, data);
	return data;
}

busdata
BankRAMDevice::Read16(uint32 addr)
{
	uint32 offset = Decoder(addr);
	accstat_read[offset >> 20] = MainbusDevice::ACC_READ;
	uint32 data = *(uint16 *)&ram[HLW(offset)];
	putlog(4, "$%08x.W -> $%04x", addr, data);
	return data;
}

busdata
BankRAMDevice::Read32(uint32 addr)
{
	uint32 offset = Decoder(addr);
	accstat_read[offset >> 20] = MainbusDevice::ACC_READ;
	uint32 data = *(uint32 *)&ram[offset];
	putlog(4, "$%08x.L -> $%08x", addr, data);
	return data;
}

busdata
BankRAMDevice::Write8(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	accstat_write[offset >> 20] = MainbusDevice::ACC_WRITE;
	ram[HLB(offset)] = data;
	putlog(3, "$%08x.B <- $%02x", addr, data);
	return 0;
}

busdata
BankRAMDevice::Write16(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	accstat_write[offset >> 20] = MainbusDevice::ACC_WRITE;
	*(uint16 *)&ram[HLW(offset)] = data;
	putlog(3, "$%08x.W <- $%04x", addr, data);
	return 0;
}

busdata
BankRAMDevice::Write32(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	accstat_write[offset >> 20] = MainbusDevice::ACC_WRITE;
	*(uint32 *)&ram[offset] = data;
	putlog(3, "$%08x.L <- $%08x", addr, data);
	return 0;
}

busdata
BankRAMDevice::Peek8(uint32 addr)
{
	uint32 offset = Decoder(addr);
	return ram[HLB(offset)];
}

bool
BankRAMDevice::Poke8(uint32 addr, uint32 data)
{
	if ((int32)data >= 0) {
		uint32 offset = Decoder(addr);
		ram[HLB(offset)] = data;
	}
	return true;
}

void
BankRAMDevice::SetPage(uint32 page_)
{
	page = page_;
	// XXX 4MB の時の折返し確認
	pageaddr = (page << 16) & (ram_size - 1);

	putlog(2, "Page <- $%02x", page);
}

void
BankRAMDevice::FetchAccStat(uint8 *acc)
{
	// R|W をマージしながらコピー。
	for (int i = 0; i < accstat_read.size(); i += 4) {
		uint32 tmp = 0;
		tmp |= *(uint32 *)&accstat_read[i];
		tmp |= *(uint32 *)&accstat_write[i];
		*(uint32 *)&acc[i] = tmp;
	}

	// 読んだのでリセット。
	std::fill(accstat_read.begin(), accstat_read.end(), 0);
	std::fill(accstat_write.begin(), accstat_write.end(), 0);
}
