//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// バスとデバイスとのインタフェース。
// バスに対し、デバイスがどのように接続されているかを記述する。

#pragma once

#if !defined(SELFTEST)
#include "device.h"
#endif

// バイトデバイスをバイト、ワード、ロングワードポートにミラー配置する共通部
template <class T, int rshift>
class BusIO_B_Base : public T
{
 protected:
	inline uint32 Decoder(uint32 addr) {
		return (addr >> rshift) & (T::NPORT - 1);
	}

 public:
	busdata Read8(uint32 addr) override {
		return T::Read(Decoder(addr));
	}

	busdata Write8(uint32 addr, uint32 data) override {
		return T::Write(Decoder(addr), data);
	}

	busdata Peek8(uint32 addr) override {
		return T::Peek(Decoder(addr));
	}

	bool Poke8(uint32 addr, uint32 data) override {
		return T::Poke(Decoder(addr), data);
	}
};

// バイトデバイスをバイトのまま配置する。
// NPORT によるミラーを行う以外はこれを使わないのと何も変わらない。
template <class T>
class BusIO_B : public BusIO_B_Base<T, 0> { };
// バイトデバイスをワードにミラー配置する。
template <class T>
class BusIO_BB : public BusIO_B_Base<T, 1> { };
// バイトデバイスをロングワードにミラー配置する。
template <class T>
class BusIO_BBBB : public BusIO_B_Base<T, 2> { };

// ワードデバイスをワードのまま配置する。
// NPORT によるミラーを行う以外はこれを使わないのと何も変わらない。
template <class T>
class BusIO_W : public T
{
 private:
	inline uint32 Decoder(uint32 addr) {
		return (addr >> 1) & (T::NPORT - 1);
	}

 public:
	// ワードポートからのバイト読み込みはワードで読んだ後必要な方を取る。
	busdata Read8(uint32 addr) override {
		if ((addr & 1) == 0) {
			busdata data = T::Read(Decoder(addr));
			data >>= 8;
			return data;
		} else {
			busdata data = T::Read(Decoder(addr));
			data &= 0xff;
			return data;
		}
	}
	busdata Read16(uint32 addr) override {
		return T::Read(Decoder(addr));
	}

	// ワードポートにバイト書き込みするとデータバスは OP3 OP3 OP3 OP3。
	busdata Write8(uint32 addr, uint32 data) override {
		data |= data << 8;
		return T::Write(Decoder(addr), data);
	}
	busdata Write16(uint32 addr, uint32 data) override {
		return T::Write(Decoder(addr), data);
	}

	busdata Peek8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return T::Peek(Decoder(addr)) >> 8;
		} else {
			return T::Peek(Decoder(addr)) & 0xff;
		}
	}
};

/* XXX: stub
class BusIO_L : public T
*/

// バイトデバイスを偶数アドレスに配置し、奇数アドレスはバスエラー。
template <class T>
class BusIO_BE : public T
{
 private:
	inline uint32 Decoder(uint32 addr) {
		return (addr >> 1) & (T::NPORT - 1);
	}

 public:
	busdata Read8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return T::Read(Decoder(addr));
		} else {
			return busdata::BusErr;
		}
	}
	busdata Read16(uint32 addr) override {
		busdata data = T::Read(Decoder(addr));
		data <<= 8;
		data |= 0xff;
		return data;
	}

	busdata Write8(uint32 addr, uint32 data) override {
		if ((addr & 1) == 0) {
			return T::Write(Decoder(addr), data);
		} else {
			return busdata::BusErr;
		}
	}
	busdata Write16(uint32 addr, uint32 data) override {
		return T::Write(Decoder(addr), data >> 8);
	}

	busdata Peek8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return T::Peek(Decoder(addr));
		} else {
			return busdata::BusErr;
		}
	}
};

// バイトデバイスを奇数アドレスに配置し、偶数アドレスはバスエラー。
template <class T>
class BusIO_EB : public T
{
 private:
	inline uint32 Decoder(uint32 addr) {
		return (addr >> 1) & (T::NPORT - 1);
	}

 public:
	busdata Read8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return busdata::BusErr;
		} else {
			return T::Read(Decoder(addr));
		}
	}
	busdata Read16(uint32 addr) override {
		busdata data = T::Read(Decoder(addr));
		data |= 0xff00;
		return data;
	}

	busdata Write8(uint32 addr, uint32 data) override {
		if ((addr & 1) == 0) {
			return busdata::BusErr;
		} else {
			return T::Write(Decoder(addr), data);
		}
	}
	busdata Write16(uint32 addr, uint32 data) override {
		return T::Write(Decoder(addr), data & 0xff);
	}

	busdata Peek8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return busdata::BusErr;
		} else {
			return T::Peek(Decoder(addr));
		}
	}
};

// バイトデバイスを奇数アドレスに配置し、偶数アドレスは非接続。
template <class T>
class BusIO_FB : public T
{
 private:
	inline uint32 Decoder(uint32 addr) {
		return (addr >> 1) & (T::NPORT - 1);
	}

 public:
	busdata Read8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return 0xff;
		} else {
			return T::Read(Decoder(addr));
		}
	}
	busdata Read16(uint32 addr) override {
		busdata data = T::Read(Decoder(addr));
		data |= 0xff00;
		return data;
	}

	busdata Write8(uint32 addr, uint32 data) override {
		if ((addr & 1) == 0) {
			return 0;
		} else {
			return T::Write(Decoder(addr), data);
		}
	}
	busdata Write16(uint32 addr, uint32 data) override {
		return T::Write(Decoder(addr), data & 0xff);
	}

	busdata Peek8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return 0xff;
		} else {
			return T::Peek(Decoder(addr));
		}
	}
};

// バイトデバイスをロングワードの先頭に配置し、他は非接続。
template <class T>
class BusIO_BFFF : public T
{
 protected:
	inline uint32 Decoder(uint32 addr) {
		return (addr >> 2) & (T::NPORT - 1);
	}

 public:
	busdata Read8(uint32 addr) override {
		if ((addr & 3) == 0) {
			return T::Read(Decoder(addr));
		} else {
			return 0xff;
		}
	}
	busdata Read16(uint32 addr) override {
		if ((addr & 2) == 0) {
			busdata data = T::Read(Decoder(addr));
			data <<= 8;
			data |= 0xff;
			return data;
		} else {
			return 0xffff;
		}
	}
	busdata Read32(uint32 addr) override {
		busdata data = T::Read(Decoder(addr));
		data <<= 24;
		data |= 0x00ffffff;
		return data;
	}

	busdata Write8(uint32 addr, uint32 data) override {
		if ((addr & 3) == 0) {
			return T::Write(Decoder(addr), data);
		} else {
			return 0;
		}
	}
	busdata Write16(uint32 addr, uint32 data) override {
		if ((addr & 2) == 0) {
			return T::Write(Decoder(addr), data >> 8);
		} else {
			return 0;
		}
	}
	busdata Write32(uint32 addr, uint32 data) override {
		return T::Write(Decoder(addr), data >> 24);
	}

	busdata Peek8(uint32 addr) override {
		if ((addr & 3) == 0) {
			return T::Peek(Decoder(addr));
		} else {
			return 0xff;
		}
	}

	bool Poke8(uint32 addr, uint32 data) override {
		if ((addr & 3) == 0) {
			return T::Poke(Decoder(addr), data);
		} else {
			// 非接続なので、書き換えようとしてもエラーも起きず何も起きない。
			return true;
		}
	}
};

// ワードデバイスをロングワード先頭に配置し、他は非接続。
template <class T>
class BusIO_WF : public T
{
 private:
	inline uint32 Decoder(uint32 addr) {
		return (addr >> 2) & (T::NPORT - 1);
	}
 public:
	busdata Read8(uint32 addr) override {
		busdata data;
		switch (addr & 3) {
		 case 0:
			data = T::Read(Decoder(addr));
			data >>= 8;
			return data;
		 case 1:
			data = T::Read(Decoder(addr));
			data &= 0xff;
			return data;
		 case 2:
		 case 3:
			return 0xff;
		}
		__unreachable();
	}
	busdata Read16(uint32 addr) override {
		if ((addr & 2) == 0) {
			return T::Read(Decoder(addr));
		} else {
			return 0xffff;
		}
	}

	// バイト書き込みならデータバスは OP3 OP3 OP3 OP3。
	busdata Write8(uint32 addr, uint32 data) override {
		if ((addr & 2) == 0) {
			data |= data << 8;
			return T::Write(Decoder(addr), data);
		} else {
			return 0;
		}
	}
	busdata Write16(uint32 addr, uint32 data) override {
		if ((addr & 2) == 0) {
			return T::Write(Decoder(addr), data);
		} else {
			return 0;
		}
	}
	busdata Peek8(uint32 addr) override {
		switch (addr & 3) {
		 case 0:
			return T::Peek(Decoder(addr)) >> 8;
		 case 1:
			return T::Peek(Decoder(addr)) & 0xff;
		 case 2:
		 case 3:
			return 0xff;
		}
		__unreachable();
	}
};

// X680x0 の SPC を収容するための専用バスアダプタ。
// ここは SASI 跡地 32バイト分と SCSI 用地 32バイト分を合わせた 64 バイト
// (0x40 バイト) でのミラーになっているため、BusIO_EB では表現できない。
// SASI 跡地は実際には読み出しパターンがあるようだが、本題ではないので
// とりあえず全部 0xff を返しておく。
template <class T>
class BusIO_X68kSPC : public T
{
 private:
	inline uint32 Decoder(uint32 addr) {
		// SASI 跡地と SCSI を合わせて 0x40 バイトでミラーしている
		return (addr >> 1) & (0x20 - 1);
	}

 public:
	busdata Read8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return busdata::BusErr;
		} else {
			addr = Decoder(addr);
			if (addr < 0x10) {
				// 前半16ポートは SASI 跡地。
				return 0xff;
			} else {
				// 後半16ポートは SCSI。
				return T::Read(addr - 0x10);
			}
		}
	}
	busdata Read16(uint32 addr) override {
		addr = Decoder(addr);
		if (addr < 0x10) {
			// 前半16ポートは SASI 跡地
			return 0xffff;
		} else {
			busdata data = T::Read(addr - 0x10);
			data |= 0xff00;
			return data;
		}
	}

	busdata Write8(uint32 addr, uint32 data) override {
		if ((addr & 1) == 0) {
			return busdata::BusErr;
		} else {
			addr = Decoder(addr);
			if (addr < 0x10) {
				// 前半16ポートは SASI 跡地。
				// XXX 書き込みはどうなる?
				return 0;
			} else {
				// 後半16ポートは SCSI。
				return T::Write(addr - 0x10, data);
			}
		}
	}
	busdata Write16(uint32 addr, uint32 data) override {
		addr = Decoder(addr);
		if (addr < 0x10) {
			// 前半16ポートは SASI 跡地。
			// XXX 書き込みはどうなる?
			return 0;
		} else {
			// 後半16ポートは SCSI。
			return T::Write(addr - 0x10, data & 0xff);
		}
	}

	busdata Peek8(uint32 addr) override {
		if ((addr & 1) == 0) {
			return busdata::BusErr;
		} else {
			addr = Decoder(addr);
			if (addr < 0x10) {
				// 前半16ポートは SASI 跡地。
				return 0xff;
			} else {
				// 後半16ポートは SCSI。
				return T::Peek(addr - 0x10);
			}
		}
	}
};
