//
// nono
// Copyright (C) 2023 nono project
// Licensed under nono-license.txt
//

//
// 拡張エリアセット
//

// 拡張エリアセットの範囲は $eaff80 〜 $eaffff までの 128バイトで、
// 他は (現状デバイスがないので) バスエラーになる。
// X68kIODevice の構造の都合上、$eae000 〜 $eaffff までの 8KB のアクセスが
// すべてここに来るので、BusIO を使わず自前でさばく必要がある。
// (本来は MuxDevice で分離すべき)

#include "extarea.h"
#include "mpu.h"
#include "x68kio.h"

// コンストラクタ
ExtAreaDevice::ExtAreaDevice()
	: inherited(OBJ_EXTAREA)
{
}

// デストラクタ
ExtAreaDevice::~ExtAreaDevice()
{
}

// 初期化
bool
ExtAreaDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	x68kio = GetX68kIODevice();

	return true;
}

// リセット
void
ExtAreaDevice::ResetHard(bool poweron)
{
	// リセット時はすべて 0 らしい。(X68030 Inside/Out)
	for (int i = 0; i < 5; i++) {
		x68kio->SetExtarea(i, 0x00);
	}
}

busdata
ExtAreaDevice::Read8(uint32 addr)
{
	if (addr >= 0xeaff80) {
		// 拡張エリアセット部分は書き込み専用で読み出しはすべてバスエラー。
		putlog(2, "Read $%06x.B (BusErr)", addr);
	}
	// 他はすべてバスエラー。
	// XXX wait?
	return busdata::BusErr;
}

busdata
ExtAreaDevice::Read16(uint32 addr)
{
	if (addr >= 0xeaff80) {
		// 拡張エリアセット部分は書き込み専用で読み出しはすべてバスエラー。
		putlog(2, "Read $%06x.W (BusErr)", addr);
	}
	// 他はすべてバスエラー。
	// XXX wait?
	return busdata::BusErr;
}

busdata
ExtAreaDevice::Write8(uint32 addr, uint32 data)
{
	busdata r;

	if (addr >= 0xeaff80) {
		// 拡張エリアセット内は16バイト単位でミラー。
		uint32 offset = addr & 0x0f;

		// 偶数アドレスはバスエラー。
		// 奇数アドレスはレジスタ部分は書き込み可能。以降はバスエラー。
		switch (offset) {
		 case 0x01:
		 case 0x03:
		 case 0x05:
		 case 0x07:
		 case 0x09:
			offset >>= 1;
			putlog(1, "$%06x [%d] <- $%02x", addr, offset, data);
			x68kio->SetExtarea(offset, data);
			r = 0;
			break;
		 default:
			putlog(2, "Write $%06x.B (BusErr)", addr);
			r.SetBusErr();
			break;
		}
	} else {
		// 他はすべてバスエラー。
		r.SetBusErr();
	}

	// XXX wait?
	return r;
}

busdata
ExtAreaDevice::Write16(uint32 addr, uint32 data)
{
	// 拡張エリアセット内は EB 構造。
	return Write8(addr + 1, data & 0xff);
}

busdata
ExtAreaDevice::Peek8(uint32 addr)
{
	return busdata::BusErr;
}
