//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// メインバス (共通部分)
//

// メインバス付近に必要な機能は次のようになっている。
//
//				Luna	News	X68k	X68kIO	XPbus
// InitMainbus	o		o		o		o		x
// FC access	o		o		o		o		x
// SwitchBoot	o		o		o		o		x
// ResetByMPU	o		o		o		*1		x	(*1:どっちでもいい)
//
// Access Map	o		o		o		x		x
// HV access	o		o		o		o		?
// Peek32		o		o		o		x		x
// Singleton	o		o		o		x		x
//
// もともと FC 付きアクセス機能を Mainbus からX68kIO に伸ばすために
// MainbusDevice を導入したが、一方モニタの登録などでは Mainbus は
// システムに複数存在してはならないため、これらは同時に成立しない。
// そのため、表の上半分の機能を MainbusBaseDevice として独立させ、
// X68kIO はここから派生。表の下半分の機能を MainbusDevice とする。
//
// XPbus はどちらも不要なので (XP プロセッサから見ればメインバスだが、
// ここの Mainbus はそういう意味ではないので) ただの IODevice。

// IODevice
//   |
//   |  +-------------------+
//   +--| MainbusBaseDevice | (InitMainbus() と FC アクセスを持つ)
//   |  +-------------------+
//   |    |
//   |    |  +---------------+
//   |    +--| MainbusDevice | (これがメインバス、システムに1つ)
//   |    |  +---------------+
//   |    |    |
//   |    |    |  +-----------------+
//   |    |    +--| Mainbus24Device | (上位8ビットがテーブルで表せるメインバス)
//   |    |    |  +-----------------+
//   |    |    |    |
//   |    |    |    +-- LunaMainbus
//   |    |    |    |    +-- Luna1Mainbus
//   |    |    |    |    +-- Luna88kMainbus
//   |    |    |    +-- NewsMainbus
//   |    |    |    +-- Virt68kMainbus
//   |    |    +-- X68kMainbus (上位8ビットが IODevice で表せないため)
//   |    |
//   |    +-- X68kIODevice
//   |
//   +-- XPbusDevice (これはただの IODevice)

#include "mainbus.h"
#include "syncer.h"

//
// メインバスの素質を持つ基本クラス
//

// コンストラクタ
MainbusBaseDevice::MainbusBaseDevice(int objid_)
	: inherited(objid_)
{
}

// デストラクタ
MainbusBaseDevice::~MainbusBaseDevice()
{
}

// MPU の RESET 命令によるリセット
void
MainbusBaseDevice::ResetByMPU()
{
}

// 必ず busaddr の方を呼ぶこと。

busdata
MainbusBaseDevice::Read8(uint32 addr)
{
	assertmsg(false, "%s should not be called", __FUNCTION__);
}

busdata
MainbusBaseDevice::Read16(uint32 addr)
{
	assertmsg(false, "%s should not be called", __FUNCTION__);
}

busdata
MainbusBaseDevice::Read32(uint32 addr)
{
	assertmsg(false, "%s should not be called", __FUNCTION__);
}

busdata
MainbusBaseDevice::Write8(uint32 addr, uint32 data)
{
	assertmsg(false, "%s should not be called", __FUNCTION__);
}

busdata
MainbusBaseDevice::Write16(uint32 addr, uint32 data)
{
	assertmsg(false, "%s should not be called", __FUNCTION__);
}

busdata
MainbusBaseDevice::Write32(uint32 addr, uint32 data)
{
	assertmsg(false, "%s should not be called", __FUNCTION__);
}

busdata
MainbusBaseDevice::ReadBurst16(uint32 addr, uint32 *dst)
{
	return busdata::BusErr;
}

busdata
MainbusBaseDevice::WriteBurst16(uint32 addr, const uint32 *src)
{
	return busdata::BusErr;
}

busdata
MainbusBaseDevice::ReadBurst16(busaddr addr, uint32 *dst)
{
	return busdata::BusErr;
}

busdata
MainbusBaseDevice::WriteBurst16(busaddr addr, const uint32 *src)
{
	return busdata::BusErr;
}

// ブートページを切り替える際の共通処理。
// 派生クラスから呼ぶこと。
void
MainbusBaseDevice::SwitchBootPage(bool isrom)
{
	// ログ表示
	if (isrom) {
		putlog(1, "SwitchBootPage ROM (Boot)");
	} else {
		putlog(1, "SwitchBootPage RAM (Normal)");
	}

	// 同期モード指示
	GetSyncer()->RequestBootPageMode(isrom);
}

// モニタ表示用にデバイス名を整形。
// 7文字に切り詰めるのと、括弧があれば取り除いて Disable 色にする。
/*static*/ std::pair<std::string, TA>
MainbusBaseDevice::FormatDevName(const Device *dev)
{
	std::string name = dev->GetName();
	TA attr;

	if (name[0] == '(') {
		// 括弧付きなら、括弧を取り除いて Disable 色で表示
		int len = std::min(7, (int)name.size() - 2);
		name = name.substr(1, len);
		attr = TA::Disable;
	} else {
		name = name.substr(0, 7);
		attr = TA::Normal;
	}

	return std::make_pair(name, attr);
}


//
// メインバス (基本クラス)
//

// コンストラクタ
MainbusDevice::MainbusDevice(int objid_)
	: inherited(objid_)
{
	monitor_accstat.func =
		ToMonitorCallback(&MainbusDevice::MonitorUpdateAccStat);
	// 表示行数が機種ごとに異なるので継承側でセットしている。
	monitor_accstat.Regist(ID_MONITOR_ACCSTAT);

	// 通常は 32bit 空間全体 (そうでない機種はコンストラクタで上書きする)
	accstat_bitlen = 32;
}

// デストラクタ
MainbusDevice::~MainbusDevice()
{
}

// 初期化
bool
MainbusDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	// アクセス状況の表示用配列。
	accstat_rw.resize(2048 >> (32 - accstat_bitlen));

	return true;
}

// リセット
void
MainbusDevice::ResetHard(bool poweron)
{
	inherited::ResetHard(poweron);

	std::fill(accstat_read.begin(), accstat_read.end(), 0);
	std::fill(accstat_write.begin(), accstat_write.end(), 0);
}

// ハイパーバイザ的読み込み
busdata
MainbusDevice::HVRead8(uint32 addr)
{
	busaddr baddr(addr, busaddr::Read | busaddr::S);
	return Read8(baddr);
}

// ハイパーバイザ的読み込み (ミスアライン可)
busdata
MainbusDevice::HVRead16(uint32 addr)
{
	busaddr baddr(addr, busaddr::Read | busaddr::S);
	if (__predict_true((addr & 1) == 0)) {
		return Read16(baddr);
	} else {
		busdata data;
		data = Read8(baddr);
		baddr += 1;
		data |= Read8(baddr);
		return data;
	}
}

// ハイパーバイザ的読み込み (ミスアライン可)
busdata
MainbusDevice::HVRead32(uint32 addr)
{
	if (__predict_true((addr & 3) == 0)) {
		busaddr baddr(addr, busaddr::Read | busaddr::S);
		return Read32(baddr);
	} else {
		busdata data;
		data = HVRead16(addr + 0);
		data <<= 16;
		data |= HVRead16(addr + 2);
		return data;
	}
}

// ハイパーバイザ的書き込み
busdata
MainbusDevice::HVWrite8(uint32 addr, uint32 data)
{
	busaddr baddr(addr, busaddr::Write | busaddr::S);
	return Write8(baddr, data);
}

// ハイパーバイザ的書き込み (ミスアライン可)
busdata
MainbusDevice::HVWrite16(uint32 addr, uint32 data)
{
	busaddr baddr(addr, busaddr::Write | busaddr::S);
	if (__predict_true((addr & 1) == 0)) {
		return Write16(baddr, data);
	} else {
		busdata r;
		r = Write8(baddr, data >> 8);
		baddr += 1;
		r |= Write8(baddr, data & 0xff);
		return r;
	}
}

// ハイパーバイザ的書き込み (ミスアライン可)
busdata
MainbusDevice::HVWrite32(uint32 addr, uint32 data)
{
	if (__predict_true((addr & 3) == 0)) {
		busaddr baddr(addr, busaddr::Write | busaddr::S);
		return Write32(baddr, data);
	} else {
		busdata r;
		r  = HVWrite16(addr + 0, data >> 16);
		r |= HVWrite16(addr + 2, data & 0xffff);
		return r;
	}
}

uint64
MainbusDevice::Peek32(uint32 addr)
{
	uint64 data;

	if (__predict_false((addr & 3) != 0))
		VMPANIC("unaligned access $%08x", addr);

	data  = Peek8(addr++) << 24;
	data |= Peek8(addr++) << 16;
	data |= Peek8(addr++) << 8;
	data |= Peek8(addr);
	return data;
}

void
MainbusDevice::MonitorUpdateAccStat(Monitor *, TextScreen& screen)
{
	std::array<char, 64 + 5> buf;
	uint32 baseaddr;

	// R|W をマージしながらコピー。
	if (__predict_true(accstat_rw.size() == accstat_read.size())) {
		// 全域。
		for (int i = 0; i < accstat_read.size(); i += 8) {
			uint64 tmp = 0;
			tmp |= *(uint64 *)&accstat_read[i];
			tmp |= *(uint64 *)&accstat_write[i];
			*(uint64 *)&accstat_rw[i] = tmp;
		}
		baseaddr = 0;
	} else {
		// NEWS なら表示に使うのは前 1/4 のみで、後ろはミラー。
		// ただしアドレス表記は $c000'0000 以降(後ろの 1/4)。
		int offset = accstat_rw.size();
		int ndiv = accstat_read.size() / offset;
		for (int i = 0, iend = accstat_rw.size(); i < iend; i += 8) {
			uint64 tmp = 0;
			for (int j = 0; j < ndiv; j++) {
				tmp |= *(uint64 *)&accstat_read[i + j * offset];
				tmp |= *(uint64 *)&accstat_write[i + j * offset];
			}
			*(uint64 *)&accstat_rw[i] = tmp;
		}
		baseaddr = 0xc0;
	}

	// 読んだのでリセット。
	std::fill(accstat_read.begin(), accstat_read.end(), 0);
	std::fill(accstat_write.begin(), accstat_write.end(), 0);

	screen.Clear();
	screen.Print(0, 0,
		"Shows %dbit space. 2MB/char. 'R':Read, 'W':Write, 'A':Read+Write",
		accstat_bitlen);
	screen.Print(12, 1,
		"+$00    +$01     +$02    +$03     +$04    +$05     +$06    +$07");

	for (int y = 0, yend = accstat_rw.size() / 64; y < yend; y++) {
		const uint8 *a = &accstat_rw[y * 64];
		uint8 avail = accstat_avail[y];
		char *d = &buf[0];
		std::fill(buf.begin(), buf.end() - 1, ' ');
		for (int i = 0; i < 8; i++) {
			// 先頭に空白を入れといて、参照時は buf+1 から参照する。
			if ((i & 1) == 0) {
				d++;
			}
			if ((int8)avail < 0) {
				// 何かしらデバイスがある。16MB、8文字分。
				for (int j = 0; j < 8; j++) {
					uint op = *a & 3;
					*d++ = ".RWA"[op];
					a++;
				}
			} else {
				// この 16MB には何のデバイスもない。
				d += 8;
				a += 8;
			}
			avail <<= 1;
		}
		*d = '\0';
		screen.Print(0, y + 2, "$%02x00'0000: %s",
			baseaddr + y * 0x08, &buf[1]);
	}
}


//
// 上位8ビットがテーブルで表せる構造のメインバス。
//

// コンストラクタ (オブジェクト名指定なし)
Mainbus24Device::Mainbus24Device()
	: MainbusDevice(OBJ_MAINBUS)
{
	monitor.func = ToMonitorCallback(&Mainbus24Device::MonitorUpdate);
	monitor.SetSize(75, 33);
	monitor.Regist(ID_MONITOR_MAINBUS);
}

Mainbus24Device::~Mainbus24Device()
{
}

// アクセス状況のエリア情報初期化。
// 継承側の InitMainbus() の終わりで呼ぶこと。
void
Mainbus24Device::InitAccStat()
{
	for (int i = 0; i < accstat_avail.size(); i++) {
		uint bits = 0;
		for (int j = 0; j < 8; j++) {
			int n = i * 8 + j;
			int id = devtable[n]->GetId();
			bits <<= 1;
			if (id != OBJ_BUSERR && id != OBJ_NOPIO) {
				bits |= 1;
			}
		}
		accstat_avail[i] = bits;
	}

	// デバッグ用
	if (0) {
		for (int i = 0; i < accstat_avail.size(); i++) {
			printf("%02x: ", i);
			uint8 a = accstat_avail[i];
			for (int j = 0; j < 8; j++) {
				if ((int8)a < 0) {
					printf("1");
				} else {
					printf("0");
				}
				a <<= 1;
			}
			printf("\n");
		}
	}
}

inline IODevice *
Mainbus24Device::Decoder(uint32 addr) const
{
	return devtable[addr >> 24];
}

#define READ(NN)	do {	\
	uint32 pa = addr.Addr();	\
	accstat_read[pa >> 21] = ACC_READ;	\
	IODevice *dev = Decoder(pa);	\
	return dev->__CONCAT(Read,NN)(pa);	\
} while (0)

#define WRITE(NN)	do {	\
	uint32 pa = addr.Addr();	\
	accstat_write[pa >> 21] = ACC_WRITE;	\
	IODevice *dev = Decoder(pa);	\
	return dev->__CONCAT(Write,NN)(pa, data);	\
} while (0)

busdata
Mainbus24Device::Read8(busaddr addr)
{
	READ(8);
}

busdata
Mainbus24Device::Read16(busaddr addr)
{
	READ(16);
}

busdata
Mainbus24Device::Read32(busaddr addr)
{
	READ(32);
}

busdata
Mainbus24Device::Write8(busaddr addr, uint32 data)
{
	WRITE(8);
}

busdata
Mainbus24Device::Write16(busaddr addr, uint32 data)
{
	WRITE(16);
}

busdata
Mainbus24Device::Write32(busaddr addr, uint32 data)
{
	WRITE(32);
}

busdata
Mainbus24Device::ReadBurst16(busaddr addr, uint32 *dst)
{
	uint32 pa = addr.Addr();
	IODevice *dev = Decoder(pa);
	busdata r = dev->ReadBurst16(pa, dst);
	if (__predict_true(r.IsBusErr() == false)) {
		accstat_read[pa >> 21] = ACC_READ;
	}
	return r;
}

busdata
Mainbus24Device::WriteBurst16(busaddr addr, const uint32 *src)
{
	uint32 pa = addr.Addr();
	IODevice *dev = Decoder(pa);
	busdata r = dev->WriteBurst16(pa, src);
	if (__predict_true(r.IsBusErr() == false)) {
		accstat_write[pa >> 21] = ACC_WRITE;
	}
	return r;
}

busdata
Mainbus24Device::Peek8(uint32 addr)
{
	IODevice *dev = Decoder(addr);
	return dev->Peek8(addr);
}

bool
Mainbus24Device::Poke8(uint32 addr, uint32 data)
{
	IODevice *dev = Decoder(addr);
	return dev->Poke8(addr, data);
}

void
Mainbus24Device::MonitorUpdate(Monitor *, TextScreen& screen)
{
	screen.Clear();

	// 012345678901234567890
	// $0000'0000: 1234567

	for (int i = 0; i < 8; i++) {
		screen.Print(12 + i * 8, 0, "+$0%x", i);
	}

	int idx = 0;
	for (int i = 0; i < devtable.size() / 8; i++) {
		screen.Print(0, i + 1, "$%02x00'0000:", idx);
		for (int j = 0; j < 8; j++) {
			auto pair = FormatDevName(devtable[idx]);
			const std::string& name = pair.first;
			TA attr = pair.second;
			screen.Print(12 + j * 8, i + 1, attr, "%s", name.c_str());
			idx++;
		}
	}
}
