//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// MPU (MC68030)
//

#include "mpu680x0.h"
#include "config.h"
#include "debugger.h"
#include "interrupt.h"
#include "m68030cache.h"
#include "mainapp.h"
#include "scheduler.h"
#include "syncer.h"
#include "uimessage.h"
#include "vectortable.h"
#include <cmath>

static Syncer *gSyncer;

// コンストラクタ
MPU680x0Device::MPU680x0Device()
	: inherited()
{
	SetName("MPU(68030)");

	// MMU.xRP.DT == 0 になってはいけない(MMU 構成例外)。
	SetSRP(0x00000001, 0);
	SetCRP(0x00000001, 0);
	// FPU
	fpu_init();
	// キャッシュ
	icache.reset(new m68030Cache());

	// レジスタモニター
	reg_monitor.func = ToMonitorCallback(&MPU680x0Device::MonitorUpdateReg);
	reg_monitor.SetSize(78, 19);
	reg_monitor.Regist(ID_MONITOR_MPUREG);

	// ATC モニター
	atc_monitor.func = ToMonitorCallback(&MPU680x0Device::MonitorUpdateATC);
	atc_monitor.SetSize(113, m68030ATCTable::LINES + 6);
	atc_monitor.Regist(ID_MONITOR_MPUATC);

	// キャッシュモニター
	cache_monitor.func = ToMonitorCallback(&MPU680x0Device::MonitorUpdateCache);
	cache_monitor.SetSize(48, 20);
	cache_monitor.Regist(ID_MONITOR_MPUCACHE);

	SetTrace(false);
}

// デストラクタ
MPU680x0Device::~MPU680x0Device()
{
}

void
MPU680x0Device::SetLogLevel(int loglevel_)
{
	inherited::SetLogLevel(loglevel_);

	atc.SetLogLevel(loglevel_);
}

// 初期化
bool
MPU680x0Device::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	debugger = GetDebugger();
	interruptdev = GetInterruptDevice();
	// static 関数から参照するため。うーん…。
	gSyncer = syncer;

	// 割り込み時のディレイ。
	//
	// 外部デバイスが 68030 の IPL0-2 をアサートしてから、68030 がそれを
	// 認識(確定)するのに 1 ~ 2 クロックかかる。
	// その後レベルを比較するのは次のアップエッジ。
	// その後 /IPEND をアサートするのはその次のアップエッジ。
	// 合わせると、IPL0-2 が変化してから /IPEND がアサートされるまでは
	// 2 ~ 3 クロックかかる。(Figure 8-4)
	intr_event.time = 3 * clock_nsec;

	// FPU (6888x)
	if (gMainApp.IsX68030()) {
		// X68k は設定による
		const ConfigItem& item = gConfig->Find("fpu-type");
		std::string type = item.AsString();
		if (type == "none" || type == "0") {
			has_fpu = 0;
		} else if (type == "68881" || type == "1") {
			has_fpu = 1;
		} else {
			item.Err();
			return false;
		}
	} else {
		// LUNA は 68881 固定 (設定は見ない)
		has_fpu = 1;
	}

	vectortable = GetVectorTable();

	// 割り込みイベントコールバック設定
	intr_event.func = ToEventCallback(&MPU680x0Device::InterruptCallback);
	intr_event.SetName(GetName() + " Interrupt");
	scheduler->RegistEvent(intr_event);

	return true;
}

// リセット
void
MPU680x0Device::ResetHard(bool poweron)
{
	if (poweron) {
		// サイクルを初期化。
		used_cycle = 0;

		// 履歴を初期化。電源(再)投入時のみ行う。
		exhist.Clear();
		brhist.Clear();

		for (int i = 0; i < 16; i++) {
			reg.R[i] = 0xcccccccc;
		}
		// XXX ただし X68030 IPLROM には A6 を初期化せずに書き込んでいるバグが
		// あるため、A6 を概ね RAM のあるあたりに指定しておく…。
		reg.A[6] = 0x000ccccc;

		// この後起きるリセット例外で PC を初期化する前に参照することになるので
		// これだけここでも初期化しておく。
		reg.pc = 0;
	}
	resetting = true;

	// リセット例外を 520 クロック後に起こす。
	// 520 は、電源オン時、Vcc が最小動作規定値に達してから最低 520 クロック
	// の間アサートしなければならないという値であって、実際これだけかかる
	// とかいう値ではない。が、こちら側の都合で、この後実行される RAM の
	// ResetHard() がブートページを用意した後で ExceptionReset() で
	// SP, PC をフェッチするという順序にしないといけないため、リセット例外が
	// 起動するまでに時間がかかるというところを都合よく真似ておく。

	exec_event.func = ToEventCallback(&MPU680x0Device::ResetCallback);
	exec_event.time = 520 * clock_nsec;
	exec_event.SetName(GetName() + " Reset");
	scheduler->RestartEvent(exec_event);
}

// リセット例外コールバック。
// MPU リセットから規定時間後に呼ばれる。
void
MPU680x0Device::ResetCallback(Event& ev)
{
	resetting = false;
	intr_pending = false;

	// リセット例外を実行
	int cycle = ExceptionReset();

	// コールバックを ResetCallback から Exec* に差し替える前のここで
	// トレース状態を初期化する。
	SetTrace(debugger->IsTrace());

	// 以降は通常走行。
	if (cpu_state == CPU_STATE_NORMAL) {
		exec_event.func = exec_short;
		exec_event.time = cycle * clock_nsec;
		exec_event.SetName(GetName() + " Execute");
		scheduler->RestartEvent(exec_event);
	} else {
		// リセット例外でダブルバスフォールトならここで停止。
	}
}

// 割り込みチェック、STOP/HALT 処理を含んだ完全な実行サイクルのコールバック。
// SR の I2:I0 が変わった直後などに呼ばれる。
// ここでは割り込みをチェックし、必要なら割り込み処理を起動、
// そうでなければ通常の命令実行を行う。
void
MPU680x0Device::ExecFull(Event& ev)
{
	if (cpu_state == CPU_STATE_HALT) {
		// HALT なら停止。
		return;
	}

	// マスクよりも割り込みレベルのほうが高ければ割り込み処理。
	// この時点で intr_pending が true の可能性もあることに留意。
	// 割り込みレベルは変化してないので Level 7 の特別処理は不要
	// (割り込みレベルの変化はここではなく InterruptCallback() で処理)。
	if (intr_level > reg.intr_mask) {
		intr_pending = true;
	}

	if (intr_pending) {
		// 割り込み処理を起動。

		intr_pending = false;

		// 前の命令実行後となるここで割り込みを起動
		ev.time = ExceptionInterrupt();
		scheduler->StartEvent(ev);
	} else {
		if (cpu_state == CPU_STATE_NORMAL) {
			// コールバックを通常処理に戻してから...
			exec_event.func = exec_short;

			// この場で次の命令を実行。
			ExecShort(ev);
		} else {
			// STOP ならここで停止。
		}
	}
}

// トレース実行
void
MPU680x0Device::ExecTrace(Event& ev)
{
	debugger->ExecMain();
	ExecFull(ev);
}

// トレース状態を設定する
void
MPU680x0Device::SetTrace(bool enable)
{
	if (enable) {
		exec_short = ToEventCallback(&MPU680x0Device::ExecTrace);
		exec_full  = ToEventCallback(&MPU680x0Device::ExecTrace);
	} else {
		exec_short = ToEventCallback(&MPU680x0Device::ExecShort);
		exec_full  = ToEventCallback(&MPU680x0Device::ExecFull);
	}
	exec_event.func = exec_full;

	// STOP 命令中にオンにするには、停止しているイベントを起こす。
	// STOP 命令中にオフにする場合は、何もしなくていい。
	if (enable && exec_event.IsRunning() == false) {
		scheduler->StartEvent(exec_event);
	}
}

// 割り込みレベルが変わったことを MPU に通知。
void
MPU680x0Device::Interrupt(int level)
{
	assertmsg(0 <= level && level <= 7, "level=%d", level);

	intr_event.code = level;
	if (intr_event.IsRunning() == false) {
		scheduler->StartEvent(intr_event);
	}
}

// 外部割り込みイベントコールバック。
// ev.code 0-7 は新しい割り込みレベル。
//
// 割り込みは本来命令境界ごとにチェックするが、実際に割り込みが起きうるのは
// 1) 割り込み信号線が変化した時
// 2) 割り込みマスクを変更した時
// しかないはずなので、この時だけ命令境界でチェックする。
// ここはその 1) のほう。その2のほうは core.cpp:SetSR() 参照。
//
// 外部割り込みが来たので、レベル比較などをして、必要なら次の命令境界で
// 割り込みチェックを行うようにするところ。
void
MPU680x0Device::InterruptCallback(Event& ev)
{
	int level = ev.code;
	int mask = reg.intr_mask;

	// 信号線が 2 クロック期間安定していること云々は省略

	// 割り込みを起こすかどうかに関わらずここで更新。
	intr_level = level;

	// ホールトならここまで。
	if (cpu_state == CPU_STATE_HALT) {
		return;
	}

	// 割り込み例外を起動するのは
	// o レベルがマスクより高い
	// o マスクに関係なく、レベルが下位から 7 に変化した場合
	//   (ここは変化した時だけ呼ばれるのと 7 より上はないので
	//   level==7 なら必ず「下位から 7 に変化」のはず)
	if (level > mask || level == 7) {
		// 通知した結果割り込みが受け付けられたので、
		// この命令終了時のコールバックを割り込み処理起動に差し替える。
		MakeNextFull();
		intr_pending = true;

		// STOP 状態だったらここでイベントを開始する
		if (cpu_state == CPU_STATE_STOP) {
			// STOP から抜ける時間?
			exec_event.time = 1 * clock_nsec;
			scheduler->StartEvent(exec_event);
		}
	}
}

// 次の命令境界は Full にする。
void
MPU680x0Device::MakeNextFull()
{
	exec_event.func = exec_full;
}

void
MPU680x0Device::ChangeState(uint32 new_state)
{
	// new_mode (CPU_STATE_*) と
	// RequestCPUMode() の引数 Syncer::SCHED_CPU_* は
	// 実は同じ値なのでそのまま渡してよいことにする。
	static_assert(CPU_STATE_NORMAL == Syncer::SCHED_CPU_NORMAL, "");
	static_assert(CPU_STATE_STOP   == Syncer::SCHED_CPU_STOP,   "");
	static_assert(CPU_STATE_HALT   == Syncer::SCHED_CPU_HALT,   "");

	if (cpu_state != new_state) {
		cpu_state = new_state;
		// スケジューラに通知
		syncer->RequestCPUMode(cpu_state);
	}
}

// A-Line 命令エミュレーションのコールバックを指定。
void
MPU680x0Device::SetALineCallback(bool (*callback)(MPU680x0Device *, void*),
	void *arg)
{
	aline_callback = callback;
	aline_arg = arg;
}

// F-Line 命令エミュレーションのコールバックを指定。
void
MPU680x0Device::SetFLineCallback(bool (*callback)(MPU680x0Device *, void*),
	void *arg)
{
	fline_callback = callback;
	fline_arg = arg;
}

// RESET 命令
void
MPU680x0Device::ops_reset()
{
	auto mainbus = GetMainbusDevice();
	mainbus->ResetByMPU();
}

// ホールト状態
void
MPU680x0Device::Halt()
{
	ChangeState(CPU_STATE_HALT);

	// この命令完了後に停止するため exec を差し替える。
	MakeNextFull();

	// UI に通知 (メッセージボックス表示とか)
	UIMessage::Post(UIMessage::HALT);
}

// モニター更新 (レジスタウィンドウ)
void
MPU680x0Device::MonitorUpdateReg(Monitor *, TextScreen& screen)
{
	m68kreg tmp;
	uint32 tmp_ppc;
	uint32 tmp_state;
	int x;
	int y;

	screen.Clear();

	// ローカルにコピー
	// ただし一部は reg ではなく cpu クラス本体にある
	tmp = reg;
	tmp_ppc = ppc;
	tmp_state = cpu_state;

	// データレジスタ、アドレスレジスタ
	for (int i = 0; i < 4; i++) {
		screen.Print(0, i, "D%d:%08x D%d:%08x  A%d:%08x A%d:%08x",
			i + 0, tmp.R[i],
			i + 4, tmp.R[i + 4],
			i + 0, tmp.R[i + 8],
			i + 4, tmp.R[i + 12]);
	}

	// 5列目
	x = 50;
	// SR
	screen.Print(x, 0, "SR:%04x(%c%c%c%c%c)",
		tmp.GetSR(),
		(tmp.ccr.IsX()) ? 'X' : '-',
		(tmp.ccr.IsN()) ? 'N' : '-',
		(tmp.ccr.IsZ()) ? 'Z' : '-',
		(tmp.ccr.IsV()) ? 'V' : '-',
		(tmp.ccr.IsC()) ? 'C' : '-');

	// *SP
	uint ms = (tmp.s ? 2 : 0) | (tmp.m ? 1 : 0); // T1 T0 S M
	if (ms == 3) {
		// Supervisor (Master) mode (A7=MSP, USP/ISP)
		screen.Print(x, 1, TA::Disable, "USP:%08x", tmp.usp);
		screen.Print(x, 2, TA::Disable, "ISP:%08x", tmp.isp);
	} else if (ms == 2) {
		// Supervisor (Interrupt) mode (A7=ISP, USP/MSP)
		screen.Print(x, 1, TA::Disable, "USP:%08x", tmp.usp);
		screen.Print(x, 2, TA::Disable, "MSP:%08x", tmp.msp);
	} else {
		// User mode (A7=USP, ISP/MSP)
		screen.Print(x, 1, TA::Disable, "ISP:%08x", tmp.isp);
		screen.Print(x, 2, TA::Disable, "MSP:%08x", tmp.msp);
	}

	// 6列目
	// PC: 01234567
	// SFC:1  DFC:1
	// VBR:01234567
	x = 66;
	screen.Print(x, 0, "PC: %08x", tmp_ppc);
	screen.Print(x, 1, "SFC:%d  DFC:%d", tmp.GetSFC(), tmp.GetDFC());
	screen.Print(x, 2, "VBR:%08x", tmp.vbr);

	switch (tmp_state) {
	 case CPU_STATE_NORMAL:
		screen.Puts(50, 3, "State: Running");
		break;
	 case CPU_STATE_STOP:
		screen.Puts(50, 3, "State: STOP instruction");
		break;
	 case CPU_STATE_HALT:
		screen.Puts(50, 3, TA::On, "State: Double Bus Fault");
		break;
	 default:
		screen.Print(50, 3, TA::On, "corrupted cpu_state=%d", tmp_state);
		break;
	}

	// Cache
	y = 4;
	screen.Puts(0, y++, "<Cache>");
	screen.Print(0, y,  "CACR:%08x(", tmp.cacr);
	screen.Puts(14, y, TA::OnOff(tmp.cacr & M68K::CACR_WA),  "WA");
	screen.Puts(17, y, TA::OnOff(tmp.cacr & M68K::CACR_DBE), "DBE");
	screen.Puts(21, y, TA::OnOff(tmp.cacr & M68K::CACR_FD),  "FD");
	screen.Puts(24, y, TA::OnOff(tmp.cacr & M68K::CACR_ED),  "ED");
	screen.Puts(27, y, TA::OnOff(tmp.cacr & M68K::CACR_IBE), "IBE");
	screen.Puts(31, y, TA::OnOff(tmp.cacr & M68K::CACR_FI),  "FI");
	screen.Puts(34, y, TA::OnOff(tmp.cacr & M68K::CACR_EI),  "EI");
	screen.Putc(36, y, ')');
	screen.Print(50, y, "CAAR:%08x", tmp.caar);
	y++;

	// MMU
	x = 0;
	screen.Puts(x, y++, "<MMU>");
	screen.Print(x, y,     "SRP:%08x_%08x", tmp.srp.h, tmp.srp.l);
	screen.Print(x, y + 1, "CRP:%08x_%08x", tmp.crp.h, tmp.crp.l);
	x += 25;
	for (int i = 0; i < 2; i++) {
		uint32 tt = tmp.tt[i];
		screen.Print(x, y + i,
			"TT%d:%08x(%c%c%c Addr=$%02x Mask=$%02x FC=%d FCMask=%d)",
			i, tt,
			(tt & m68030TT::E)   ? 'E' : '-',
			(tt & m68030TT::CI)  ? 'C' : '-',
			(tt & m68030TT::RWM) ? '-' : ((tt & m68030TT::RW)  ? 'R' : 'W'),
			(tt >> 24),
			(tt >> 16) & 0xff,
			(tt >>  4) & 0x07,
			(tt      ) & 0x07);
	}
	y += 2;
	screen.Print(0, y,
		"TC:%08x(%c%c%c IS=$%x TIA-D=%x:%x:%x:%x PS=$%x)",
		tmp.tc,
		(tmp.tc & m68030TC::TC_E)   ? 'E' : '-',
		(tmp.tc & m68030TC::TC_SRE) ? 'S' : '-',
		(tmp.tc & m68030TC::TC_FCL) ? 'F' : '-',
		(tmp.tc >> 16) & 0xf,
		(tmp.tc >> 12) & 0xf,
		(tmp.tc >>  8) & 0xf,
		(tmp.tc >>  4) & 0xf,
		(tmp.tc      ) & 0xf,
		(tmp.tc >> 20) & 0xf);
	x = 49;
	screen.Print(x, y, "MMUSR:%04x(", tmp.mmusr);
	x += 11;
	screen.Putc(x++, y, TA::OnOff(tmp.mmusr & m68030MMUSR::B) | 'B');
	screen.Putc(x++, y, TA::OnOff(tmp.mmusr & m68030MMUSR::L) | 'L');
	screen.Putc(x++, y, TA::OnOff(tmp.mmusr & m68030MMUSR::S) | 'S');
	screen.Putc(x++, y, '-');
	screen.Putc(x++, y, TA::OnOff(tmp.mmusr & m68030MMUSR::W) | 'W');
	screen.Putc(x++, y, TA::OnOff(tmp.mmusr & m68030MMUSR::I) | 'I');
	screen.Putc(x++, y, TA::OnOff(tmp.mmusr & m68030MMUSR::M) | 'M');
	screen.Puts(x, y, "- -");
	x += 3;
	screen.Putc(x++, y, TA::OnOff(tmp.mmusr & m68030MMUSR::T) | 'T');
	screen.Print(x, y, "---N=%d)", tmp.mmusr & m68030MMUSR::N);
	y++;

	// FPU
	x = 0;
	screen.Puts(x, y++, "<FPU>");
	for (int i = 0; i < 8; i++) {
		screen.Print(x, y + i, "FP%d:%04x_%08x_%08x (%-20s)",
			i,
			tmp.fpframe.fpf_regs[i * 3 + 0] >> 16,
			tmp.fpframe.fpf_regs[i * 3 + 1],
			tmp.fpframe.fpf_regs[i * 3 + 2],
			"not yet");
	}

	x = 51;

	// FPCR
	static const char * const rpstr[] = {
		".EXT",
		".SGL",
		".DBL",
		".???",
	};
	uint32 fpcr = tmp.fpframe.fpf_fpcr;
	screen.Print(x, y++, "FPCR:%04x", fpcr);
	screen.Puts(x +  2, y, TA::OnOff(fpcr & 0x8000), "BS");
	screen.Puts(x +  5, y, TA::OnOff(fpcr & 0x4000), "SN");
	screen.Puts(x +  8, y, TA::OnOff(fpcr & 0x2000), "OP");
	screen.Puts(x + 11, y, TA::OnOff(fpcr & 0x1000), "OV");
	screen.Puts(x + 14, y, TA::OnOff(fpcr & 0x0800), "UF");
	screen.Puts(x + 17, y, TA::OnOff(fpcr & 0x0400), "DZ");
	screen.Puts(x + 20, y, TA::OnOff(fpcr & 0x0200), "I2");
	screen.Puts(x + 23, y, TA::OnOff(fpcr & 0x0100), "I1");
	y++;
	screen.Print(x + 2, y++, "RP=%s RM=%s",
		rpstr[(fpcr >> 6) & 3],
		rmstr[(fpcr >> 4) & 3]);

	// FPSR
	uint32 fpsr = tmp.fpframe.fpf_fpsr;
	uint32 cc = fpsr >> 24;
	screen.Print(x, y++, "FPSR:%08x", fpsr);
	screen.Puts(x +  2, y, TA::OnOff(cc & 0x08), "N");
	screen.Puts(x +  4, y, TA::OnOff(cc & 0x04), "Z");
	screen.Puts(x +  6, y, TA::OnOff(cc & 0x02), "Inf");
	screen.Puts(x + 10, y, TA::OnOff(cc & 0x01), "NAN");
	screen.Print(x + 14, y, "Q=$%02x", (fpsr >> 16) & 0xff);
	y++;
	screen.Puts(x +  2, y, TA::OnOff(fpsr & 0x8000), "BS");
	screen.Puts(x +  5, y, TA::OnOff(fpsr & 0x4000), "SN");
	screen.Puts(x +  8, y, TA::OnOff(fpsr & 0x2000), "OP");
	screen.Puts(x + 11, y, TA::OnOff(fpsr & 0x1000), "OV");
	screen.Puts(x + 14, y, TA::OnOff(fpsr & 0x0800), "UF");
	screen.Puts(x + 17, y, TA::OnOff(fpsr & 0x0400), "DZ");
	screen.Puts(x + 20, y, TA::OnOff(fpsr & 0x0200), "I2");
	screen.Puts(x + 23, y, TA::OnOff(fpsr & 0x0100), "I1");
	y++;
	screen.Puts(x +  2, y, TA::OnOff(fpsr & 0x80), "IOP");
	screen.Puts(x +  6, y, TA::OnOff(fpsr & 0x80), "OVFL");
	screen.Puts(x + 11, y, TA::OnOff(fpsr & 0x80), "UNFL");
	screen.Puts(x + 16, y, TA::OnOff(fpsr & 0x80), "DZ");
	screen.Puts(x + 19, y, TA::OnOff(fpsr & 0x80), "INEX");
	y++;

	// FPIAR
	screen.Print(x, y++, "FPIAR:%08x", tmp.fpframe.fpf_fpiar);
}

static int
sort_atc(const void *a, const void *b)
{
	const uint32 *ua = (const uint32 *)a;
	const uint32 *ub = (const uint32 *)b;
	return *ub - *ua;
}


// モニター更新 (ATC)
void
MPU680x0Device::MonitorUpdateATC(Monitor *, TextScreen& screen)
{
	m68030ATCTable *table;

	screen.Clear();

	struct {
		int fc;
		const char *name;
	} fclist[] = {
		{ 1,	"User/Data" },
		{ 2,	"User/Program" },
		{ 5,	"Super/Data" },
		{ 6,	"Super/Program" },
	};

	for (int t = 0; t < countof(fclist); t++) {
		int x = t * 29;
		int fc = fclist[t].fc;
		const char *name = fclist[t].name;

		table = atc.fctables[fc];
		screen.Print(x, 0, "FC=%d %s", fc, name);
		screen.Puts(x, 1, "No. LAddr    PAddr    Flag");

		std::array<uint32, m68030ATCTable::LINES * 2> sbuf;
		for (int i = 0; i < m68030ATCTable::LINES; i++) {
			sbuf[i * 2 + 0] = table->line[i].age;
			sbuf[i * 2 + 1] = i;
		}
		qsort(&sbuf[0], m68030ATCTable::LINES, sizeof(uint32) * 2, sort_atc);

		// エントリ表示
		for (int i = 0; i < m68030ATCTable::LINES; i++) {
			int idx = sbuf[i * 2 + 1];
			m68030ATCLine& a = table->line[idx];
			TA attr;
			if (a.IsInvalid()) {
				attr = TA::Disable;
			} else {
				attr = TA::Normal;
			}

			screen.Print(x, i + 2, attr, "[%x] %08x %08x %c%c%c%c",
				idx,
				a.GetLAddr(),
				a.GetPAddr(),
				a.IsBusError() ? 'B' : '-',
				a.IsCInhibit() ? 'C' : '-',
				a.IsWProtect() ? 'P' : '-',
				a.IsModified() ? 'M' : '-');
		}

		// ヒット率とミス率。
		// atc_total を先にインクリメントしているので、
		// 細かいことはともかくロック取らなくても死にはしなさそう。
		uint64 tthit = table->atc_tthit;
		uint64 miss  = table->atc_miss;
		uint64 total = table->atc_total;
		int y = m68030ATCTable::LINES + 2;
		screen.Puts(x + 1, y, "TT Match");
		if (tthit == 0) {
			screen.Puts(x + 12, y, "0");
		} else {
			screen.Print(x + 10, y, "%5.1f%%", (double)(tthit * 100) / total);
		}
		y++;

		screen.Puts(x + 1, y, "ATC Match");
		if (total == 0) {
			screen.Puts(x + 12, y, "0");
		} else {
			screen.Print(x + 10, y, "%5.1f%%",
				(double)((total - miss - tthit) * 100) / total);
		}
		y++;

		screen.Puts(x + 1, y, "ATC Miss");
		if (miss == 0) {
			screen.Puts(x + 12, y, "0");
		} else {
			screen.Print(x + 10, y, "%5.1f%%", (double)(miss * 100) / total);
		}
		y++;
	}

	screen.Print(0, m68030ATCTable::LINES + 5,
		"Note: The real 68030 ATC is single FC-mixed 22-entries. "
		"This ATC with FC-separated 16-entries is by nono.");
}

// モニタ更新 (キャッシュ)
void
MPU680x0Device::MonitorUpdateCache(Monitor *, TextScreen& screen)
{
	TA attr;
	int y;

	screen.Clear();

	// <I-Cache>
	// 01234567890123456789012345678901234567890123456789
	// Tag Address  Data
	// S.012345'00: 00000000 00000000 -------- 00000000

	screen.Print(0, 0, "<I-Cache>");
	screen.Print(0, 1, "Tag Address  Data");

	attr = icache_enable ? TA::Normal : TA::Disable;

	y = 2;
	for (int idx = 0; idx < icache->line.size(); idx++) {
		auto& l = icache->line[idx];
		screen.Print(0, y, attr, "%c.%06x'%x0:",
			(l.IsSuper() ? 'S' : 'U'),
			(l.TagAddr() >> 8),
			idx);

		for (int i = 0; i < 4; i++) {
			if (l.valid[i]) {
				screen.Print(13 + 9 * i, y, attr, "%08x", l.data[i]);
			} else {
				screen.Print(13 + 9 * i, y, TA::Disable, "--------");
			}
		}
		y++;
	}

	y++;
	screen.Print(0, y, "<D-Cache> Not Implemented");
}
