//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// スプライトコントローラ
//

#include "sprite.h"
#include "mpu.h"

// InsideOut p.135
// 書き込みは VRAM とレジスタで異なる。
// 読み込みは VRAM、レジスタとも 17 ウェイト。
const busdata vram_wait = busdata::Wait(18 * 40_nsec);
const busdata reg_wait  = busdata::Wait(20 * 40_nsec);
const busdata read_wait = busdata::Wait(17 * 40_nsec);

// コンストラクタ
SpriteDevice::SpriteDevice()
	: inherited(OBJ_SPRITE)
{
	mem.reset(new uint8[4 * 8192]);
}

// デストラクタ
SpriteDevice::~SpriteDevice()
{
}

// XXX ResetHard() 未調査

busdata
SpriteDevice::Read8(uint32 addr)
{
	busdata data;

	if (addr >= 0xeb8000) {
		data = mem[HB(addr - 0xeb8000)];
	} else {
		VMPANIC("not impl");
	}

	data |= read_wait;
	return data;
}

busdata
SpriteDevice::Read16(uint32 addr)
{
	busdata data;

	if (addr >= 0xeb8000) {
		data = *(uint16 *)&mem[addr - 0xeb8000];

	} else if (addr < 0xeb0400) {
		int n = (addr - 0xeb0000) / 8;
		switch (addr % 8) {
		 case 0:
			data = reg.sprite[n].xpos;
			break;
		 case 2:
			data = reg.sprite[n].ypos;
			break;
		 case 4:
			data = reg.sprite[n].col;
			break;
		 case 6:
			data = reg.sprite[n].prw;
			break;
		 default:
			__unreachable();
		}
	} else if (0xeb0800 <= addr && addr < 0xeb0812) {
		switch (addr) {
		 case 0xeb0800:
			data = reg.bg0x;
			break;
		 case 0xeb0802:
			data = reg.bg0y;
			break;
		 case 0xeb0804:
			data = reg.bg1x;
			break;
		 case 0xeb0806:
			data = reg.bg1y;
			break;
		 case 0xeb0808:
			data = reg.bgctrl;
			break;
		 case 0xeb080a:
			data = reg.htotal;
			break;
		 case 0xeb080c:
			data = reg.hdisp;
			break;
		 case 0xeb080e:
			data = reg.vdisp;
			break;
		 case 0xeb0810:
			data = reg.res;
			break;
		 default:
			__unreachable();
		}
	} else {
		data.SetBusErr();
	}

	data |= read_wait;
	return data;
}

busdata
SpriteDevice::Write8(uint32 addr, uint32 data)
{
	if (addr >= 0xeb8000) {
		mem[HB(addr - 0xeb8000)] = data;
		return vram_wait;
	} else {
		VMPANIC("not impl");
		return reg_wait;
	}
}

busdata
SpriteDevice::Write16(uint32 addr, uint32 data)
{
	putlog(2, "$%06x <- $%04x", addr, data);
	if (addr >= 0xeb8000) {
		*(uint16 *)&mem[addr - 0xeb8000] = data;
		return vram_wait;
	}
	if (addr < 0xeb0400) {
		int n = (addr - 0xeb0000) / 8;
		switch (addr % 8) {
		 case 0:
			reg.sprite[n].xpos = data & 0x3ff;
			break;
		 case 2:
			reg.sprite[n].ypos = data & 0x3ff;
			break;
		 case 4:
			reg.sprite[n].col = data & 0xcfff;
			break;
		 case 6:
			reg.sprite[n].prw = data & 0x3;
			break;
		}
		return reg_wait;
	}
	if (0xeb0800 <= addr && addr < 0xeb0812) {
		switch (addr) {
		 case 0xeb0800:
			reg.bg0x = data & 0x3ff;
			break;
		 case 0xeb0802:
			reg.bg0y = data & 0x3ff;
			break;
		 case 0xeb0804:
			reg.bg1x = data & 0x3ff;
			break;
		 case 0xeb0806:
			reg.bg1y = data & 0x3ff;
			break;
		 case 0xeb0808:
			reg.bgctrl = data & 0x23f;
			break;
		 case 0xeb080a:
			reg.htotal = data & 0xff;
			break;
		 case 0xeb080c:
			reg.hdisp = data & 0x3f;
			break;
		 case 0xeb080e:
			reg.vdisp = data & 0xff;
			break;
		 case 0xeb0810:
			reg.res = data & 0x1f;
			break;
		}
		return reg_wait;
	}

	busdata r = reg_wait;	// ?
	r.SetBusErr();
	return r;
}

busdata
SpriteDevice::Peek8(uint32 addr)
{
	if (addr >= 0xeb8000) {
		return mem[HB(addr - 0xeb8000)];
	}
	if (addr < 0xeb0400) {
		int n = (addr - 0xeb0000) / 8;
		switch (addr % 8) {
		 case 0:
			return reg.sprite[n].xpos;
		 case 2:
			return reg.sprite[n].ypos;
		 case 4:
			return reg.sprite[n].col;
		 case 6:
			return reg.sprite[n].prw;
		}
	}
	if (0xeb0800 <= addr && addr < 0xeb0812) {
		switch (addr) {
		 case 0xeb0800:
			return reg.bg0x;
		 case 0xeb0802:
			return reg.bg0y;
		 case 0xeb0804:
			return reg.bg1x;
		 case 0xeb0806:
			return reg.bg1y;
		 case 0xeb0808:
			return reg.bgctrl;
		 case 0xeb080a:
			return reg.htotal;
		 case 0xeb080c:
			return reg.hdisp;
		 case 0xeb080e:
			return reg.vdisp;
		 case 0xeb0810:
			return reg.res;
		}
	}

	return busdata::BusErr;
}
