//
// nono
// Copyright (C) 2024 nono project
// Licensed under nono-license.txt
//

//
// virt68k の I/O 空間担当
//

// メモリマップ。
// NetBSD/virt68k は $ff00'0000 以降が I/O 空間で、デバイス配置は
// ブートローダ (実質 BIOS) からカーネルにパラメータとして渡される。
// 各デバイスの先頭番地こそ自由だが、同種のデバイスが連続する場合
// 配置間隔がデバイスによって異なっておりその長さはハードコードのようだ。
// GF* は 0x1000 番地ずつ、VirtIO は 0x200 番地ずつとなっている。
//
// メモリマップはこんな感じにする。
//
// $ff00'0000-$ff00'ffff PROM
// $ff01'xxxx            Goldfish Interrupt(PIC) #1..#6
// $ff01'1000-$ff01'1fff  Lv1
// $ff01'2000-$ff01'2fff  Lv2
// $ff01'3000-$ff01'3fff  Lv3
// $ff01'4000-$ff01'4fff  Lv4
// $ff01'5000-$ff01'5fff  Lv5
// $ff01'6000-$ff01'6fff  Lv6
// $ff02'xxxx            Goldfish RTC+Timer
// $ff02'0000-$ff02'0fff  Timer
// $ff02'1000-$ff02'1fff  RTC
// $ff03'0000-$ff03'0fff Goldfish TTY
// $ff04'0000-$ff04'0fff Qemu Virtual System Controller
// $ff07'xxxx            VirtIO
//
// VirtIO は間隔が違うので最後に置く。
// ここは 1スロット $200 で、最大 127個まで置けるようだ。
// デバイスが存在しないスロットはバスエラーではなく、DEVICE_ID = INVALID
// で応答しなければならない。
//
// $ff07'0000-$ff07'01ff  VirtIO(00)
// $ff07'0200-$ff07'03ff  VirtIO(01)
//  :
// $ff07'fe00-$ff07'ffff  VirtIO(7f)

#include "v68kio.h"
#include "config.h"
#include "goldfish_pic.h"
#include "goldfish_rtc.h"
#include "goldfish_timer.h"
#include "goldfish_tty.h"
#include "mainbus.h"
#include "qemusysctlr.h"
#include "romemu_virt68k.h"
#include "virtio_base.h"
#include "virtio_block.h"
#include "virtio_entropy.h"
#include "virtio_net.h"
#include "vm.h"

// コンストラクタ
V68kIODevice::V68kIODevice()
	: inherited(OBJ_V68KIO)
{
	NEWDV(PROM, new Virt68kROMEmuDevice());
	NEWDV(GFPIC1, new GFPICDevice(1));
	NEWDV(GFPIC2, new GFPICDevice(2));
	NEWDV(GFPIC3, new GFPICDevice(3));
	NEWDV(GFPIC4, new GFPICDevice(4));
	NEWDV(GFPIC5, new GFPICDevice(5));
	NEWDV(GFPIC6, new GFPICDevice(6));
	NEWDV(GFRTC, new GFRTCDevice());				// required by GFTimer
	NEWDV(GFTimer, new GFTimerDevice());
	NEWDV(GFTTY, new GFTTYDevice());
	NEWDV(QemuSysCtlr, new QemuSysCtlrDevice());
	NEWDV(VirtIONone, new VirtIONoneDevice());

	buserr = GetBusErrDevice();

	// まず全域をバスエラーで埋める。
	for (int i = 0; i < table.size(); i++) {
		table[i] = buserr;
	}

	// 少ないので手動で埋めていく。
	for (int i = 0; i < 0x10; i++) {
		table[i] = pPROM.get();
	}
	table[0x11] = pGFPIC1.get();
	table[0x12] = pGFPIC2.get();
	table[0x13] = pGFPIC3.get();
	table[0x14] = pGFPIC4.get();
	table[0x15] = pGFPIC5.get();
	table[0x16] = pGFPIC6.get();
	table[0x20] = pGFTimer.get();
	table[0x21] = pGFRTC.get();
	table[0x30] = pGFTTY.get();
	table[0x40] = pQemuSysCtlr.get();

	// VirtIO 用は、設定によるので後で代入する。
	// ここはバスエラーではない。
	for (int i = 0; i < viotable.size(); i++) {
		viotable[i] = pVirtIONone.get();
	}
}

// デストラクタ
V68kIODevice::~V68kIODevice()
{
}

// 動的なコンストラクション
bool
V68kIODevice::Create()
{
	int slot;

	// (QEMU virt68k の?)
	// virtio はスロットによって割り込み先を固定してあるようだ。
	// スロット  0 .. 31 は PIC2(Lv2) の IRQ1..32、
	// スロット 32 .. 63 は PIC3(Lv3) の IRQ1..32、
	// スロット 64 .. 95 は PIC4(Lv4) の IRQ1..32、
	// スロット 96 ..127 は PIC5(Lv5) の IRQ1..32 となる。
	//
	// VirtIO* はとりあえず先例にならって Lv5 につなぐことにしてみる。
	slot = 96;

	// Net
	if (true) {
		pVirtIO[slot].reset(new VirtIONetDevice(slot));
		if ((bool)pVirtIO[slot] == false) {
			return false;
		}
		viotable[slot] = pVirtIO[slot].get();
		slot++;
	}

	// Entropy
	if (true) {
		pVirtIO[slot].reset(new VirtIOEntropyDevice(slot));
		if ((bool)pVirtIO[slot] == false) {
			return false;
		}
		viotable[slot] = pVirtIO[slot].get();
		slot++;
	}

	// Block (とりあえず8個まで)
	for (int id = 0; id < 8; id++) {
		const std::string key = string_format("virtio-block%d-image", id);
		const ConfigItem& item = gConfig->Find(key);
		const std::string& image = item.AsString();

		// 空ならデバイスなし
		if (image.empty()) {
			// 未接続部分のパラメータは減らしておくか。
			// --show-config するとさすがに無駄に多くて邪魔なので。
			gConfig->Delete(string_format("virtio-block%d-writeignore", id));
			continue;
		}

		pVirtIO[slot].reset(new VirtIOBlockDevice(slot, id));
		if ((bool)pVirtIO[slot] == false) {
			item.Err();	// ?
			return false;
		}
		viotable[slot] = pVirtIO[slot].get();
		slot++;
	}

	return true;
}

bool
V68kIODevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	return true;
}

inline IODevice *
V68kIODevice::SearchDevice(uint32 addr) const
{
	if (__predict_true(addr < 0xff07'0000)) {
		// FF0X'X000
		uint32 idx = (addr >> 12) & 0x7f;
		return table[idx];
	} else if (__predict_true(addr < 0xff08'0000)) {
		// FF07'XX00
		uint32 idx = (addr >> 9) & 0x7f;
		return viotable[idx];
	} else {
		return buserr;
	}
}

busdata
V68kIODevice::Read8(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Read8(addr);
}

busdata
V68kIODevice::Read16(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Read16(addr);
}

busdata
V68kIODevice::Read32(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Read32(addr);
}

busdata
V68kIODevice::Write8(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	return d->Write8(addr, data);
}

busdata
V68kIODevice::Write16(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	return d->Write16(addr, data);
}

busdata
V68kIODevice::Write32(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	return d->Write32(addr, data);
}

busdata
V68kIODevice::Peek8(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Peek8(addr);
}

bool
V68kIODevice::Poke8(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	return d->Poke8(addr, data);
}
