//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// ビデオコントローラ
//

#pragma once

#include "device.h"
#include "bitmap.h"
#include "color.h"
#include "event.h"
#include "planevram.h"
#include <array>

class Renderer;

class VideoCtlrDevice : public IODevice
{
	using inherited = IODevice;

	static const uint32 baseaddr = 0xe82000;

 public:
	VideoCtlrDevice();
	~VideoCtlrDevice() override;

	bool Init() override;
	void ResetHard(bool poweron) override;

	busdata Read8(uint32 addr) override;
	busdata Read16(uint32 addr) override;
	busdata Write8(uint32 addr, uint32 data) override;
	busdata Write16(uint32 addr, uint32 data) override;
	busdata Peek8(uint32 addr) override;
	bool Poke8(uint32 addr, uint32 data) override;

	// コントラストを取得 (Peek() からも呼ぶので副作用を起こしてはいけない)
	int GetContrast() const { return (int)(target_contrast / 0x11); }
	// コントラストを設定
	void SetContrast(int);

	// 画面合成
	bool Render(BitmapRGBX& dst, ModifyInfo&);

	// 加工済みテキストパレットの先頭アドレスを取得
	// レンダラとかからの参照用
	const std::vector<ColorXBGR>& GetHostPalette() const { return hostcolor; }

	// 生パレットを uint32 形式にしたものを返す (GUI 用)
	const std::vector<uint32> GetIntPalette() const;

 private:
	inline uint32 Decoder(uint32 addr) const;
	inline uint Offset2NR(uint32 offset) const;

	uint32 GetNR(uint reg) const;
	void SetNR(uint reg, uint32 data);
	void MakePalette();
	void RenderContrast(BitmapRGBX& dst, const BitmapRGBX& src);
	static void RenderContrast_gen(BitmapRGBX&, const BitmapRGBX&, uint32);
#if defined(HAVE_AVX2)
	static void RenderContrast_avx2(BitmapRGBX&, const BitmapRGBX&, uint32);
	bool enable_avx2 {};
#endif

	void ContrastCallback(Event& ev);

	// ワードレジスタ R0, R1, R2 は内部では nr0 .. nr5 というバイトレジスタ
	// として扱う。
	std::array<uint8, 6> nreg {};

	// テキスト画面
	BitmapRGBX textbmp {};

	// メモリを確保する時のサイズ計算のため型を uint8 にしているが、
	// 運用はワードでホストバイトオーダー。
	std::array<uint8, 1024> palette {};

	// ホスト形式のテキストパレット (レンダラ用に加工したもの)
	std::vector<ColorXBGR> hostcolor {};

	// 設定したコントラスト値 (0-255)
	uint32 target_contrast {};
	// 現在の (遷移中かもしれない) コントラスト値 (0-255)
	uint32 running_contrast {};
	// 設定を変える時の初期値 (0-255)
	uint32 initial_contrast {};
	// 設定を変える時の起点
	uint64 contrast_time0 {};

	// レンダリングに使っているコントラスト値 (0-255)
	uint32 rendering_contrast {};

	Event contrast_event { this };

	PlaneVRAMDevice *planevram {};
	Renderer *renderer {};
};

static inline VideoCtlrDevice *GetVideoCtlrDevice() {
	return Object::GetObject<VideoCtlrDevice>(OBJ_VIDEOCTLR);
}
